//! Misc utilities used in different places in the crate.

#[cfg(feature = "benchmarks")]
#[cfg_attr(docsrs, doc(cfg(feature = "benchmarks")))]
pub mod bench_util;

pub mod macros;
pub mod total_ord;
